/*{{{  includes*/
#include <stdio.h>
#include <errno.h>
#include <conc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "inmos.h"
#include "isocket.h"
#include "t9defs.h"
#include "queue.h"
#include "iserver.h"
#define	ACE_COFF
#include <filehdr.h>
#include <aouthdr.h>
#include <scnhdr.h>
#include <reloc.h>
#include <sgs.h>
/*}}}*/

/*{{{  defines and globals*/

#define	BOOT_BUFFER_SIZE	(1024*100)
unsigned char BootBuffer[BOOT_BUFFER_SIZE];

/*  COFF parameters that are to be sent to the secondary bootstrap
*/
struct coff_params {
	unsigned long TextStart;
	unsigned long TextSize;
	unsigned long DataStart;
	unsigned long DataSize;
	unsigned long BssStart;
	unsigned long BssSize;
	unsigned long NumSections;
	unsigned long Entry;
	unsigned long HeapSize;
	unsigned long StackSize;
	unsigned long DebugFlag;
}CoffParams;

#undef DEBUG
#ifdef DEBUG
#define debug_msg fprintf
#else
static debug_msg(){}
#endif

/*}}}*/

/*{{{  utilities*/

short _htons(short in)
{ short out;
  char *inptr=(char *)&in, *outptr=(char *)&out;
  outptr[0]=inptr[1];
  outptr[1]=inptr[0];
  return(out);
}

long _htonl(long in)
{ long out;
  char *inptr=(char *)&in, *outptr=(char *)&out;
  outptr[0]=inptr[3];
  outptr[1]=inptr[2];
  outptr[2]=inptr[1];
  outptr[3]=inptr[0];
  return(out);
}
/*}}}*/

/*{{{  Relocate*/
/*
 *------------------------------------------------------------------------
 *
 *	Relocate :
 *		Relocates a part of the section which has been read into
 *		BootBuffer.
 *	Parameters :
 *		RelocBuf : Relocation entries
 *		StartOffset : Offset from which to relocate
 *		EndOffset : Offset until which relocation is done
 *		FirstIndex : Pointer to relocation info in RelocBuf
 *		nReloc : Pointer to number of relocation entries.
 *		LoadAddress : Address to which relocation is performed.
 *
 *	Return value :
 *		RelPtr is set to the file offset of the next relocation
 *		entry. nReloc is set to the number of relocation entries
 *		left. The function does not return anything.
 *
 *------------------------------------------------------------------------
 */
static void Relocate( RelocBuf, StartOffset, EndOffset, nReloc, LoadAddress )
char **RelocBuf;
unsigned long StartOffset;
unsigned long EndOffset;
int *nReloc;
unsigned long LoadAddress;
{
	int i;
	struct reloc reloc;
	unsigned long address;
	int nRelocOver;
	char *RelPtr;

	nRelocOver = 0;
	RelPtr = *RelocBuf;
	for ( i = 0 ; i < *nReloc; i++, RelPtr += RELSZ ){
		blockmove( &reloc, RelPtr, RELSZ );
                reloc.r_vaddr = _htonl( reloc.r_vaddr );
                reloc.r_symndx = _htonl( reloc.r_symndx );
                reloc.r_type = _htons( reloc.r_type );
		if ( reloc.r_type != RSEG_LONG )
			continue;
		if ( reloc.r_vaddr < StartOffset ){
			fprintf(stderr, "Internal error: Unexpected reloc entry addr %x, off %x", reloc.r_vaddr,StartOffset );exit(-1);
		}
		if ( reloc.r_vaddr >= EndOffset )
			break;
		address = *(unsigned long *)(&BootBuffer[reloc.r_vaddr-StartOffset]);
		address  += LoadAddress;
		*(unsigned long *)(&BootBuffer[reloc.r_vaddr-StartOffset]) = address;
		nRelocOver++;
	}
	*RelocBuf = RelPtr;
	*nReloc -= nRelocOver;
}

/*}}}*/

/*{{{  LoadSections*/
/*
 *------------------------------------------------------------------------
 *
 *	LoadSections :
 *		Loads the sections of a COFF file at the specified address
 *		over a virtual or physical channel.
 *		CoffParameters are sent first. Each section is preceded by 
 *		its SectionInfo structure.
 *	Parameters :
 *		fp : File pointer to open COFF file
 *		NumSections : Number of sections in the COFF file
 *		LoadAddress : Address at which the file is loaded
 *			be sent
 *		ServerInfo : structure containing information about server
 *
 *	Returns nothing.
 *
 *------------------------------------------------------------------------
 */
static void LoadSections( CoffFileName, CoffFileBuf, NumSections, LoaderOutChan, LoaderInChan )
char *CoffFileBuf;
int NumSections;
Channel *LoaderOutChan, *LoaderInChan;
{
	int i, j;
	unsigned long LoadAddress=0;
	char *SectionHeaderBuf;
	SCNHDR SectionHeader;
	struct section_info {
		unsigned long SectionStart;
		unsigned long SectionSize;
	}SectionInfo;
	int ops_res;
	int buf_size, section_size;
	int nreloc;
	unsigned long relptr, StartOffset, EndOffset;
	unsigned long MemRequired=0;
	char *RelocBuf, *scnptr;

	debug_msg(stderr,  "Loading %d sections\n", NumSections );
	SectionHeaderBuf = (char *)(CoffFileBuf+FILHSZ+AOUTHSZ);
	for ( i = 0; i < NumSections; i++ ){
		blockmove( &SectionHeader, SectionHeaderBuf + i*SCNHSZ, SCNHSZ );
		SectionHeader.s_size = _htonl( SectionHeader.s_size );
		SectionHeader.s_vaddr = _htonl( SectionHeader.s_vaddr );
		if ( strcmp( SectionHeader.s_name, ".text" ) == 0 ){
			CoffParams.TextStart = SectionHeader.s_vaddr;
			debug_msg(stderr, "TextStart %x\n",  CoffParams.TextStart );
			CoffParams.TextSize = SectionHeader.s_size;
		}
		if ( strcmp( SectionHeader.s_name, ".data" ) == 0 ){
			CoffParams.DataStart = SectionHeader.s_vaddr;
			CoffParams.DataSize = SectionHeader.s_size;
		}
		if ( strcmp( SectionHeader.s_name, ".bss" ) == 0 ){
			CoffParams.BssStart = SectionHeader.s_vaddr;
			CoffParams.BssSize = SectionHeader.s_size;
		}
	}
	MemRequired = CoffParams.TextSize+CoffParams.DataSize+CoffParams.BssSize+CoffParams.HeapSize+CoffParams.StackSize;
	debug_msg(stderr, "MemRequired %d\n", MemRequired );
	_VChanOut( LoaderOutChan, &MemRequired, sizeof(MemRequired));
	_VChanIn( LoaderInChan, &LoadAddress, sizeof(unsigned long), -1);
	if ( LoadAddress == 0x80000000 ){
		fprintf(stderr, "Out of memory, cannot load task\n");exit(-1);
	}
	debug_msg(stderr, "MemStart = %x\n", LoadAddress );
	CoffParams.TextStart += LoadAddress;
	CoffParams.DataStart += LoadAddress;
	CoffParams.BssStart += LoadAddress;
	CoffParams.Entry += LoadAddress;
	debug_msg(stderr, "Writing Coff params size %d\n", sizeof(CoffParams) );
	_VChanOut( LoaderOutChan, &CoffParams, sizeof(CoffParams));
	debug_msg(stderr, "Wrote Coff params entry %x\n", CoffParams.Entry );
	for ( i = 0; i < NumSections; i++ ){
		blockmove( &SectionHeader, SectionHeaderBuf + i*SCNHSZ, SCNHSZ );
		SectionHeader.s_size = _htonl( SectionHeader.s_size );
		SectionHeader.s_paddr = _htonl( SectionHeader.s_paddr );
		SectionHeader.s_vaddr = _htonl( SectionHeader.s_vaddr );
		SectionHeader.s_scnptr = _htonl( SectionHeader.s_scnptr );
		SectionHeader.s_type = _htons( SectionHeader.s_type );
		SectionHeader.s_relptr = _htonl( SectionHeader.s_relptr );
		SectionHeader.s_nreloc = _htons( SectionHeader.s_nreloc );
		debug_msg(stderr, "Loading section %s\n", SectionHeader.s_name );
		if ( SectionHeader.s_type & STYP_CLEAR ){
			if ( strcmp(SectionHeader.s_name, ".bss" ) == 0 )
				continue;
		}
		SectionInfo.SectionStart = SectionHeader.s_vaddr + LoadAddress;
		SectionInfo.SectionSize = SectionHeader.s_size;
		debug_msg(stderr, "Writing section info\n" );
		_VChanOut( LoaderOutChan, &SectionInfo, sizeof(SectionInfo));
		debug_msg(stderr, "Wrote section info\n" );
		section_size = SectionHeader.s_size;
		StartOffset = EndOffset = SectionHeader.s_vaddr;
		relptr = SectionHeader.s_relptr;
		nreloc = SectionHeader.s_nreloc;
		if (( nreloc > 0 )&&( strcmp( SectionHeader.s_name, ".data" ) == 0 )) {
			RelocBuf = ( char *)(CoffFileBuf+SectionHeader.s_relptr);
		}
		scnptr = (char *)(CoffFileBuf + SectionHeader.s_scnptr);
		while ( section_size > 0 ){
			buf_size = ( section_size > BOOT_BUFFER_SIZE )?BOOT_BUFFER_SIZE:section_size;
			EndOffset += buf_size;
			section_size -= buf_size;
			if (!( SectionHeader.s_type & STYP_CLEAR )){
				blockmove( (char *)BootBuffer, scnptr, buf_size );
				scnptr += buf_size;
			}
			if (( nreloc > 0 )&&( strcmp( SectionHeader.s_name, ".data" ) == 0 )) {
				Relocate( &RelocBuf, StartOffset, EndOffset, &nreloc, LoadAddress );
			}
			debug_msg( stderr,  "Writing section %s size %d\n", SectionHeader.s_name, buf_size );
			_VChanOut(LoaderOutChan, BootBuffer, (unsigned long) buf_size );
			debug_msg( stderr,   "Wrote section %s size %d\n", SectionHeader.s_name, buf_size );

			StartOffset += buf_size;
		}
	}
}
/*}}}*/

/*{{{  CoffLoad*/
/*  Load a COFF file at	the specified address
*/
/*
 *------------------------------------------------------------------------
 *
 *	CoffLoad :
 *		Loads the sections of a COFF file, starting at LoadAddress. 
 *		CoffParameters are sent first. Each section is preceded by 
 *		its SectionInfo structure.
 *	Parameters :
 *		CoffFileName : Name of COFF file
 *		ServerInfo : structure containing information about server
 *
 *	Returns nothing.
 *
 *------------------------------------------------------------------------
 */
static char *CoffFileBuf=NULL;
static char RecentlyLoadedFileName[100]={0};
void CoffLoadT9000( 
char *CoffFileName,
Channel *LoaderOutChan, 
Channel *LoaderInChan, 
int AppHeapSize, int AppStackSize,
int DebugApplicationSwitch,
char *DebuggerCommand )
{
	int CoffFile;
	int res;
	FILHDR *FileHeader;
	AOUTHDR *AoutHeader;
	unsigned long length;
	unsigned long temp;
	struct stat statBuf;
	int t1, t2, t3=0, t4=0;


	t1 = ProcTime();
	if ( strcmp(CoffFileName, RecentlyLoadedFileName) != 0 ){
		strcpy( RecentlyLoadedFileName, CoffFileName );
		if (stat(CoffFileName, &statBuf) == -1) {
			fprintf(stderr, "Boot: cannot find file %s", CoffFileName);
			return;
		}
		if ( CoffFileBuf != NULL )free(CoffFileBuf);
		CoffFileBuf = malloc( statBuf.st_size );
		if ( CoffFileBuf == NULL ){
			fprintf( stderr, "Could not allocate memory to read file %s\n", CoffFileName ); return;
		}
		CoffFile = open( CoffFileName, O_RDONLY, 0 );
		if ( CoffFile < 0 ){
			fprintf(stderr, "Boot: cannot find file \"%s\"", CoffFileName);
			return;
		}
		t3 = ProcTime();
		if (( res = read( CoffFile, CoffFileBuf,  statBuf.st_size )) != statBuf.st_size ){
			fprintf(stderr,  "%s: Could not read coff file, res %d\n", CoffFileName, res );return;
		}
		close( CoffFile );
		t4 = ProcTime();
		FileHeader = (FILHDR *)CoffFileBuf;
		FileHeader->f_magic = _htons(FileHeader->f_magic);
		FileHeader->f_nscns = _htons(FileHeader->f_nscns);
		FileHeader->f_opthdr = _htons(FileHeader->f_opthdr);
		FileHeader->f_flags = _htons(FileHeader->f_flags);
		if (!ISACEMAGIC(FileHeader->f_magic)){
			fprintf(stderr,  "%s: Bad magic %x\n", CoffFileName, FileHeader->f_magic );return;
		}
		if (!( FileHeader->f_flags & F_EXEC )){
			fprintf(stderr,  "%s: Not an executable file flags %x\n" , CoffFileName, FileHeader->f_flags );return;
		}
		AoutHeader = (AOUTHDR *)(CoffFileBuf+FILHSZ);
		AoutHeader->entry = _htonl(AoutHeader->entry);
	}
	else {
		FileHeader = (FILHDR *)CoffFileBuf;
		AoutHeader = (AOUTHDR *)(CoffFileBuf+FILHSZ);
	}
	CoffParams.Entry = AoutHeader->entry;
	CoffParams.NumSections = FileHeader->f_nscns;
	CoffParams.HeapSize = AppHeapSize;
	CoffParams.StackSize = AppStackSize;
	CoffParams.DebugFlag = (unsigned long)DebugApplicationSwitch;
	LoadSections( CoffFileName, CoffFileBuf, (int)FileHeader->f_nscns, LoaderOutChan, LoaderInChan );
	debug_msg(stderr, "DebugSwitch %d Heap %d Stack %d\n",  DebugApplicationSwitch, AppHeapSize, AppStackSize );
	if ( DebugApplicationSwitch ){
		length = strlen( DebuggerCommand )+1;
		_VChanOut( LoaderOutChan, &length, sizeof(unsigned long));
		_VChanOut( LoaderOutChan, DebuggerCommand, length);
	}
	debug_msg(stderr, "Boot COFF file over\n" );
	t2 = ProcTime();
#if 0
	fprintf( stderr, "Time %g %g\n", (double)(t2-t1)*64.0/1000000.0, (double)(t4-t3)*64.0/1000000.0 );
#endif
}
/*}}}*/
