#include <stdio.h>
#include <conc.h>
#include <node.h>
#include <string.h>
#include "t9defs.h"
#include "queue.h"


int MyNodeId;
int MyTaskId;
unsigned int MessageId = 1;
RECEIVE_REQUEST *ReceiveRequestQ=NULL;
MESSAGE *MessagesReceived=NULL;
MESSAGE_HEADER *HeaderReceived;
int ReceiverChanNum;
_Semaphore SenderSemaphore[MAX_NODES][MAX_TASKS];
_Semaphore StartSemaphore, ReadySemaphore, EndSemaphore;
PVMD_PARAMETERS *pvmd_info=NULL;
int NUM_NODES;
Channel *MIN_PVM_INTERNAL_CHANNEL;
RECEIVE_REQUEST ReceiveRequestBuffer[MAX_RECEIVE_REQUESTS], *FreeReceiveRequests = NULL;
MESSAGE HeaderBuffer[MAX_MESSAGES_QUEUED], *FreeHeaders = NULL;
int NumPacketsRequired = 0;
WDESC_TYPE FreeHeaderWait = NOT_PROCESS, FreePacketWait = NOT_PROCESS;
PACKET_BUFFER ReceiveBuffer[MAX_PACKET_BUFFERS];
RECEIVE_BUFFER *FreeBuffers = NULL;
Channel **VirtualChannelList = NULL;
WDESC_TYPE PvmReceiver= NOT_PROCESS;

extern void _SemWait(_Semaphore *);
extern void _SemSignal(_Semaphore *);



#define IsEXIT( header ) ( header.flag == EXIT_MESSAGE )
#define IsSTOP( header ) ( header.flag == STOP_MESSAGE )


#define INIT_SEMAPHORE( sema ){					\
	sema.sBack = sema.sFront = NOT_PROCESS; 		\
	sema.sCount=1;						\
}

	


#define MATCHES( pattern, item ) (( pattern == -1 )||( pattern == item ))

#define FORALL( Q, element ) for ( element = Q; element != NULL; element = QNext(element))


#define MEMCPY( dest, src, len ) blockmove( dest, src, len )



/*{{{  OutChannel*/
/*
 *------------------------------------------------------------------------
 *
 *      OutChannel :
 *              Returns the channel address to communicate to another task
 *      Parameters:
 *		NodeId : Destination node
 *		TaskId : Destination task
 *
 *      Return value :
 *              Channel connecting this task to task number <TaskId> on 
 *		the node <NodeId>
 *
 *------------------------------------------------------------------------
 */
#define OutChannel( VChan, NodeId, TaskId ) {			\
	int index;						\
	if ( NodeId == MyNodeId ){				\
		VChan= (Channel *)((unsigned long)MIN_PVM_INTERNAL_CHANNEL + ((TaskId * MAX_TASKS + MyTaskId) * BYTES_PER_INTERNAL_CHANNEL ));	\
	} else {						\
		index = ( NodeId > MyNodeId )?NodeId-1:NodeId;	\
		VChan= (Channel *)((unsigned long)MIN_PVM_VIRTUAL_CHANNEL + ((MyTaskId * (NUM_NODES-1) * MAX_TASKS + index * MAX_TASKS + TaskId) * BYTES_PER_VIRTUAL_CHANNEL ) + 4);						\
	}							\
}
/*}}}*/


/*{{{  GetReceiveRequest*/
/*
 *------------------------------------------------------------------------
 *
 *      GetReceiveRequest :
 *              Checks if a request has been posted for a message that has
 *		arrived.
 *      Parameters:
 *		NodeId, TaskId	: Source of message
 *		Header		: Message header
 *		ReceiveReq	: Pointer to return the request
 *		
 *      Return value :
 *              ReceiveReq is set to the first posted receive request if any.
 *		Otherwise it is set to NULL
 *
 *------------------------------------------------------------------------
 */
#define GetReceiveRequest( NodeId, TaskId, Header, ReceiveReq ) {	\
	FORALL( ReceiveRequestQ, ReceiveReq ){				\
		if (MATCHES( ReceiveReq->SourceNode, NodeId )&&		\
		    MATCHES( ReceiveReq->SourceTask, TaskId )&&		\
		    MATCHES( ReceiveReq->tag, Header->tag )){		\
			QRemove( ReceiveRequestQ, ReceiveReq );		\
			break;						\
		}							\
	}								\
}
/*}}}*/


void *MEM_ALLOCATE( int size )
{
	void *ptr;
	if ( size == 0 )return((void *)0);
	ptr = (void *)malloc(size);
	if ( ptr == NULL ){
		char buf[50];
		sprintf( buf, "Malloc %d bytes failed \n", size );
		write( 2, buf, strlen(buf)+1);
	}
	return( ptr );
}

/*{{{  ProcSetPri*/
/*
 *------------------------------------------------------------------------
 *
 *      ProcSetPri :
 *              Sets the priority of the current thread to the specified
 *              priority.
 *
 *      Return value :
 *              Current priority
 *
 *------------------------------------------------------------------------
 */
PRIORITY_TYPE ProcSetPri( PRIORITY_TYPE priority )
{
        int CurrentPri;
        ProcGetPri( &CurrentPri );
        if ( CurrentPri == priority )
                return( (PRIORITY_TYPE)CurrentPri );
        if ( priority == High )
                ProcToHigh();
        else
                ProcToLow();
        return( (PRIORITY_TYPE)CurrentPri );
}
/*}}}*/
/*{{{  PvmReceiverThread*/
/*
 *------------------------------------------------------------------------
 *
 *      PvmReceiverThread :
 *		Receives all the messages to this task. 
 *		If a request has been posted, the message is received into
 *		the user's buffer; else the message is received and queued.
 *	Parameters :
 *		VirtualChanList : List of channels on which to wait for messages
 *
 *      Returns nothing
 *
 *------------------------------------------------------------------------
 */
void PvmReceiverThread( Channel **VirtualChanList )
{
	MESSAGE *CurrentMessage;
	RECEIVE_REQUEST *tempReceiveReq;
	RECEIVE_BUFFER *tempBuffer, *CurBuf;
	int i, index;
	int SourceNode, SourceTask;
	int len, NumBuffers;
	RECEIVE_INFO *info;

	DeQueue( FreeHeaders, CurrentMessage );
	while ( True ) {
		_SemSignal( &StartSemaphore );
		_SemWait( &ReadySemaphore );
		SourceNode = ReceiverChanNum/MAX_TASKS;
		SourceTask = ReceiverChanNum%MAX_TASKS;
		GetReceiveRequest( SourceNode, SourceTask, HeaderReceived, tempReceiveReq );
		if ( tempReceiveReq != NULL ){
			len = (tempReceiveReq->BufferSize < HeaderReceived->MessageLength )?tempReceiveReq->BufferSize:HeaderReceived->MessageLength;
#ifndef PACKETIZE
			_VChanIn( VirtualChanList[ReceiverChanNum], tempReceiveReq->BufferPtr, len );
#else
		{
			char *buf; int BytesLeft, RecvLen;
			BytesLeft = HeaderReceived->MessageLength;
			RecvLen = len;
			buf = tempReceiveReq->BufferPtr;
			if ( BytesLeft == 0 )
				_VChanIn( VirtualChanList[ReceiverChanNum], buf, 0);
			while ( BytesLeft > 0 ){
				_VChanIn( VirtualChanList[ReceiverChanNum], buf, RecvLen < PACKET_SIZE?RecvLen:PACKET_SIZE );
				RecvLen -= PACKET_SIZE; if ( RecvLen < 0 )RecvLen = 0;
				BytesLeft -= PACKET_SIZE; buf += PACKET_SIZE;
			}
		}
#endif
			if ( tempReceiveReq->info != NULL ) {
				info = tempReceiveReq->info;
				info->tag = HeaderReceived->tag;
				info->MessageLength = len;
				info->SourceNode = SourceNode;
				info->SourceTask = SourceTask;
			}
			ProcAwaken( tempReceiveReq->ReceiverProcess );
		}
		else {
			if ( HeaderReceived->MessageLength == 0 ){
				NumBuffers = 0; 
				tempBuffer = NULL;
			} else {
				NumBuffers = HeaderReceived->MessageLength/PACKET_SIZE;
				if ( HeaderReceived->MessageLength % PACKET_SIZE > 0 )NumBuffers++;
				DeQueueN( FreeBuffers, NumBuffers, tempBuffer );
				if ( tempBuffer == NULL ){
					tempBuffer = ( RECEIVE_BUFFER *)malloc( (HeaderReceived->MessageLength+EXTRA_BYTES+sizeof(RECEIVE_BUFFER)) * sizeof(char));
					if (( tempBuffer == NULL )&&(NumBuffers < MAX_PACKET_BUFFERS )){
						NumPacketsRequired = NumBuffers;
						ProcSleep(&FreePacketWait);
						DeQueueN( FreeBuffers, NumBuffers, tempBuffer );
					} else 
					{
						tempBuffer->AllocFlag = True;
						NumBuffers = 1;
					}
				}
				if ( tempBuffer == NULL )HeaderReceived->MessageLength = 0;
			}
#ifndef PACKETIZE
			_VChanIn( VirtualChanList[ReceiverChanNum], tempBuffer->Data, HeaderReceived->MessageLength);
#else
		{
			char *buf; int BytesLeft;
			BytesLeft = HeaderReceived->MessageLength;
			if ( BytesLeft == 0 )
				_VChanIn( VirtualChanList[ReceiverChanNum], &len, 0);
			else {
				CurBuf = tempBuffer;
				buf = CurBuf->Data;
			}
			while ( BytesLeft > 0 ){
				_VChanIn( VirtualChanList[ReceiverChanNum], buf, BytesLeft < PACKET_SIZE?BytesLeft:PACKET_SIZE );
				BytesLeft -= PACKET_SIZE; 
				if ( tempBuffer->AllocFlag )
					buf += PACKET_SIZE;
				else{
					CurBuf = QNext(CurBuf);
					buf = CurBuf->Data;
				}
			}
		}
#endif
			GetReceiveRequest( SourceNode, SourceTask, HeaderReceived, tempReceiveReq );
			if ( tempReceiveReq != NULL ){
				len = (tempReceiveReq->BufferSize < HeaderReceived->MessageLength )?tempReceiveReq->BufferSize:HeaderReceived->MessageLength;
				if ( tempReceiveReq->info != NULL ) {
					info = tempReceiveReq->info;
					info->tag = HeaderReceived->tag;
					info->MessageLength = len;
					info->SourceNode = SourceNode;
					info->SourceTask = SourceTask;
				}
				if ( tempBuffer != NULL ){
					if ( tempBuffer->AllocFlag ){
						MEMCPY( tempReceiveReq->BufferPtr, tempBuffer->Data, len );
						free( tempBuffer );
					} else {
						CurBuf = tempBuffer;
						while ( len > 0 ){
							char *Buffer = tempReceiveReq->BufferPtr;
							MEMCPY( Buffer, CurBuf->Data, len>PACKET_SIZE?PACKET_SIZE:len );
							CurBuf = QNext(CurBuf);
							len -= PACKET_SIZE;
							Buffer += PACKET_SIZE;
						}
						AddListN( FreeBuffers, tempBuffer );
					}
				}
				ProcAwaken( tempReceiveReq->ReceiverProcess );
			}
			else {
				CurrentMessage->Header = *HeaderReceived;
				CurrentMessage->Buffer = tempBuffer;
				CurrentMessage->SourceNode = SourceNode;
				CurrentMessage->SourceTask = SourceTask;
				EnQueue( MessagesReceived, CurrentMessage );
				DeQueue( FreeHeaders, CurrentMessage );
				if ( CurrentMessage == NULL ){
					CurrentMessage = (MESSAGE*)MEM_ALLOCATE( sizeof(MESSAGE));
					if ( CurrentMessage == NULL ){
						ProcSleep( &FreeHeaderWait );
						DeQueue( FreeHeaders, CurrentMessage );
					}
					else
						CurrentMessage->AllocFlag = True;
				}
				if ( PvmReceiver != NOT_PROCESS ){
					ProcAwaken( PvmReceiver );	
					PvmReceiver = NOT_PROCESS;
				}
			}
		}
		_SemSignal( &EndSemaphore );
	}
}
/*}}}*/


ReceiverThread( Channel *VirtualChan, int ChannelIndex )
{
	MESSAGE_HEADER Header;
	
	while ( True ){
		_VChanIn( VirtualChan, &Header, sizeof(MESSAGE_HEADER ));
		_SemWait( &StartSemaphore );
		ReceiverChanNum = ChannelIndex;
		HeaderReceived = &Header;
		_SemSignal( &ReadySemaphore );
		_SemWait( &EndSemaphore );
	}
}


PvmdReceiverThread( Channel *VirtualChan, int ChannelIndex )
{
	MESSAGE_HEADER Header;
	
	while ( True ){
		_VChanIn( VirtualChan, &Header, sizeof(MESSAGE_HEADER ));
		if ( IsSTOP( Header )){
			ResetAllChannels();
			ThreadStop();
		}
		if ( IsEXIT( Header )){
			__Exit(0);
			ResetAllChannels();
			ThreadStop();
		}
		_SemWait( &StartSemaphore );
		ReceiverChanNum = ChannelIndex;
		HeaderReceived = &Header;
		_SemSignal( &ReadySemaphore );
		_SemWait( &EndSemaphore );
	}
}


/*{{{  _init_pvm*/
/*
 *------------------------------------------------------------------------
 *
 *      _init_pvm :
 *              Initializes all data structures, and creates receiver threads
 *		to receive messages.
 *
 *      Returns nothing
 *
 *------------------------------------------------------------------------
 */
void _init_pvm()
{
	int i, j, index;
	int ChanIndex = 0;
	Channel **ChanList=NULL;
	Process *pid;
	int *ParameterBlock;
	int *MDL, *ThreadStack;

	MyNodeId = _Node.id;
	MyTaskId = _Node.process->id;
	NUM_NODES = *(char *)_get_param(1);
	MIN_PVM_INTERNAL_CHANNEL = *(Channel **)_get_param(2);
	for ( i = 0; i < NUM_NODES; i++ ){
		for ( j = 0; j < MAX_TASKS; j++ ){
			INIT_SEMAPHORE( SenderSemaphore[i][j] );
		}
	}
	INIT_SEMAPHORE( StartSemaphore );
	INIT_SEMAPHORE( ReadySemaphore );
	INIT_SEMAPHORE( EndSemaphore );
	_SemWait( &StartSemaphore );
	_SemWait( &ReadySemaphore );
	_SemWait( &EndSemaphore );
	for ( i = 0; i < MAX_RECEIVE_REQUESTS; i++ )
		AddList( FreeReceiveRequests, &ReceiveRequestBuffer[i] );
	for ( i = 0; i < MAX_MESSAGES_QUEUED; i++ ){
		AddList( FreeHeaders, &HeaderBuffer[i] );
		HeaderBuffer[i].AllocFlag = False;
	}
	for ( i = 0; i < MAX_PACKET_BUFFERS; i++ ){
		AddList( FreeBuffers, &ReceiveBuffer[i] );
		ReceiveBuffer[i].AllocFlag = False;
	}
	ChanList = (Channel **)MEM_ALLOCATE((NUM_NODES * MAX_TASKS+1)*sizeof(Channel *));
	ChanIndex = 0;
	for ( i = 0; i < NUM_NODES; i++ ){
		if ( i == MyNodeId ){
			unsigned long CurChan;
			CurChan = ((unsigned long)MIN_PVM_INTERNAL_CHANNEL + (MyTaskId * MAX_TASKS * BYTES_PER_INTERNAL_CHANNEL ));
			for ( j = 0; j < MAX_TASKS; j++, ChanIndex++, CurChan += BYTES_PER_INTERNAL_CHANNEL )
				ChanList[ChanIndex] = (Channel *)CurChan;
		} else {
			index = ( i > MyNodeId )?i-1:i;
			ChanList[ChanIndex++]= (Channel *)((unsigned long)MIN_PVM_VIRTUAL_CHANNEL + ((MyTaskId * (NUM_NODES-1) * MAX_TASKS + index * MAX_TASKS) * BYTES_PER_VIRTUAL_CHANNEL ) );
			for ( j = 1; j < MAX_TASKS; j++, ChanIndex++ ){
				ChanList[ChanIndex] = (Channel *)((unsigned long)ChanList[ChanIndex-1] + BYTES_PER_VIRTUAL_CHANNEL);
			}
		}
	}
	ChanList[ChanIndex] = NULL;
	VirtualChannelList = ChanList;
	for ( i = 0; i < ChanIndex; i++ ){

		ParameterBlock = (int *)MEM_ALLOCATE( 3 * sizeof(int));
		ParameterBlock[0] = 2 * sizeof(long);
		ParameterBlock[1] = ChanList[i];
		ParameterBlock[2] = i;

		ThreadStack = (int *)MEM_ALLOCATE( DEFAULT_STACKSIZE );
		ThreadStack += (DEFAULT_STACKSIZE/sizeof(int) - 10 );
		ThreadStack[0] = 0;
		Get_MDL(&(ThreadStack[1]));
		ThreadStack[2] = (int)&(ParameterBlock[1]);
		ThreadStack[3] = 0;
		if ( i == 0 )
			ThreadRunHigh( ThreadStack, PvmdReceiverThread );
		else
			ThreadRunHigh( ThreadStack, ReceiverThread );
	}

	ParameterBlock = (int *)MEM_ALLOCATE( 2 * sizeof(int));
	ParameterBlock[0] = 1 * sizeof(long);
	ParameterBlock[1] = ChanList;

	ThreadStack = (int *)MEM_ALLOCATE( DEFAULT_PVM_STACKSIZE );
	ThreadStack += (DEFAULT_PVM_STACKSIZE/sizeof(int) - 10 );	
	ThreadStack[0] = 0;
	Get_MDL(&(ThreadStack[1]));
	ThreadStack[2] = (int)&(ParameterBlock[1]);
	ThreadStack[3] = 0;
	ThreadRunHigh( ThreadStack, PvmReceiverThread );
}
/*}}}*/

/*{{{  _async_send*/
/*
 *------------------------------------------------------------------------
 *
 *      _async_send :
 *		Asynchronous send
 *	Parameters :
 *		tag : Message tag ( as passed by PVM )
 *		Message : Message buffer
 *		MessageLen : Length of message in bytes
 *		DestinationNode, DestinationTask : destination of message
 *
 *      Return value :
 *              Message identifier
 *
 *------------------------------------------------------------------------
 */
int _async_send( int tag, char *Message, int MessageLen, int DestinationNode, int DestinationTask )
{
	MESSAGE_HEADER header;
	Channel *VChan;
	PRIORITY_TYPE CurrentPri;

	header.MessageLength = MessageLen;
	header.tag = tag;
	OutChannel( VChan, DestinationNode, DestinationTask );
	header.flag = ASYNC_MESSAGE;
	_SemWait( &SenderSemaphore[DestinationNode][DestinationTask] );
	_VChanOut( VChan, &header, FIXUP_LEN(sizeof(header)));
#ifndef PACKETIZE
	_VChanOut( VChan, Message, FIXUP_LEN(MessageLen) );
#else
	if ( MessageLen == 0 ){
		_VChanOut( VChan, Message, 0);
	}
	while ( MessageLen > 0 ) {
		_VChanOut( VChan, Message, MessageLen>PACKET_SIZE?PACKET_SIZE:MessageLen);
		MessageLen -= PACKET_SIZE;
		Message += PACKET_SIZE;
	}
#endif
	_SemSignal( &SenderSemaphore[DestinationNode][DestinationTask] );
	return( 1 );
}
/*}}}*/



/*{{{  _blocked_receive*/
/*
 *------------------------------------------------------------------------
 *
 *      _blocked_receive :
 *              Receive a message
 *	Parameters :
 *		SourceNode, SourceTask : Id of source, -1 indicates any source
 *		tag : Message tag, -1 indicates any tag
 *		Buffer : Buffer to receive buffer
 *		BufferSize : Maximum length of message to be received
 *		info : buffer to return information on the received message.
 *
 *      Return value :
 *              0 for Success, negative value for failure
 *		info contains the message source, tag and length
 *
 *------------------------------------------------------------------------
 */
int _blocked_receive( int SourceNode, int SourceTask, int tag, char *Buffer, int BufferSize, RECEIVE_INFO *info )
{
	MESSAGE *tempMessage;
	PRIORITY_TYPE CurrentPriority;
	RECEIVE_REQUEST *tempRequest;
	RECEIVE_BUFFER *tempBuffer, *CurBuf;
	Boolean ReqAllocated = False;
	int len;
	PRIORITY_TYPE CurrentPri;


	CurrentPri = ProcSetPri( High );
	FORALL( MessagesReceived, tempMessage ){
		if ( MATCHES( SourceNode, tempMessage->SourceNode ) &&
		     MATCHES( SourceTask, tempMessage->SourceTask ) &&
		     MATCHES( tag, tempMessage->Header.tag ))
				break;
	}
	if ( tempMessage != NULL ){
		QRemove( MessagesReceived, tempMessage );
		len = BufferSize < tempMessage->Header.MessageLength?BufferSize:tempMessage->Header.MessageLength;
		if ( info != NULL ){
			info->tag = tempMessage->Header.tag;
			info->MessageLength = len;
			info->SourceNode = tempMessage->SourceNode;
			info->SourceTask = tempMessage->SourceTask;
		}
		tempBuffer = tempMessage->Buffer;
		if ( tempBuffer != NULL ){
			if ( tempBuffer->AllocFlag ){
				MEMCPY( Buffer, tempBuffer->Data, len );
				free( tempMessage->Buffer );
			}
			else {
				CurBuf = tempBuffer;
				while ( len > 0 ){
					MEMCPY( Buffer, CurBuf->Data, len>PACKET_SIZE?PACKET_SIZE:len );
					CurBuf = QNext(CurBuf);
					len -= PACKET_SIZE;
					Buffer += PACKET_SIZE;
				}
				AddListN( FreeBuffers, tempBuffer );
				if ( FreePacketWait != NOT_PROCESS ){
					int len;
					ListLength( FreeBuffers, len );
					if ( len >= NumPacketsRequired ){
						ProcAwaken( FreePacketWait );
						FreePacketWait = NOT_PROCESS;
					}
				}
			}
		}
		if ( tempMessage->AllocFlag )
			free( tempMessage );
		else {
			AddList( FreeHeaders, tempMessage );
			if ( FreeHeaderWait != NOT_PROCESS ){
				ProcAwaken( FreeHeaderWait );
				FreeHeaderWait = NOT_PROCESS;
			}
		}
	} else {
		DeQueue( FreeReceiveRequests, tempRequest );
		if ( tempRequest == NULL ){
			tempRequest = ( RECEIVE_REQUEST *)MEM_ALLOCATE( sizeof(RECEIVE_REQUEST));
			ReqAllocated = True;
		}
		tempRequest->BufferSize = BufferSize;
		tempRequest->BufferPtr = Buffer;
		tempRequest->tag = tag;
		tempRequest->SourceNode = SourceNode;
		tempRequest->SourceTask = SourceTask;
		tempRequest->info = info;
		/* Enqueue should be protected by semaphores, if not done at
		   high priority */
		EnQueue( ReceiveRequestQ, tempRequest );
		ProcSleep( &tempRequest->ReceiverProcess );
		if ( ReqAllocated )
			free( tempRequest );
		else
			AddList( FreeReceiveRequests, tempRequest );
	}	

	ProcSetPri( CurrentPri );
	return( SUCCESS );
}
/*}}}*/

int _WaitForMessage()
{
	MESSAGE *tempMessage;
	PRIORITY_TYPE CurrentPriority;
	RECEIVE_REQUEST *tempRequest;
	RECEIVE_BUFFER *tempBuffer, *CurBuf;
	Boolean ReqAllocated = False;
	int len;
	PRIORITY_TYPE CurrentPri;
	int t1, t2;

	CurrentPri = ProcSetPri( High );
	if ( !QEmpty( MessagesReceived )){
		ProcSetPri( CurrentPri );
		return(1);
	}
	ProcSleep( &PvmReceiver );
	ProcSetPri( CurrentPri );
	return( 1 );
}

/*{{{  _async_wait*/
/*
 *------------------------------------------------------------------------
 *
 *      _async_wait :
 *		Wait until a message transfer is complete
 *	Parameters :
 *		mid : message identifier
 *
 *      Returns nothing
 *
 *------------------------------------------------------------------------
 */
void _async_wait( int mid )
{
	return;
}
/*}}}*/

/*{{{  _async_done*/
/*
 *------------------------------------------------------------------------
 *
 *      _async_done :
 *		Checks if a message transfer is complete
 *	Parameters :
 *		mid : message identifier
 *
 *      Return value :
 *              1 if transfer complete, 0 otherwise.
 *
 *------------------------------------------------------------------------
 */
int _async_done( mid )
{
		return(1);
}
/*}}}*/

/*{{{  _infocount*/
/*
 *------------------------------------------------------------------------
 *
 *      _infocount :
 *              Returns the length of the first queued message
 *
 *      Return value :
 *              Length of first message
 *
 *------------------------------------------------------------------------
 */
int _infocount()
{
	int len;
	PRIORITY_TYPE CurrentPri;

	CurrentPri = ProcSetPri( High );
	if ( QEmpty(MessagesReceived) ) 
		len =  -1;
	else
		len = ( MessagesReceived->Header.MessageLength );
	ProcSetPri( CurrentPri );
	return(len);
}
/*}}}*/

/*{{{  _infonode*/
/*
 *------------------------------------------------------------------------
 *
 *      _infonode :
 *              Returns the sender of the first queued message
 *
 *      Return value :
 *              Source Nodeid  of first message
 *
 *------------------------------------------------------------------------
 */
int _infonode()
{
	if ( QEmpty(MessagesReceived) ) 
		return -1;
	else
		return( MessagesReceived->SourceNode );
}
/*}}}*/

/*{{{  _infotype*/
/*
 *------------------------------------------------------------------------
 *
 *      _infotype :
 *              Returns the tag of the first queued message
 *
 *      Return value :
 *              Tag of first message
 *
 *------------------------------------------------------------------------
 */
int _infotype()
{
	if ( QEmpty(MessagesReceived) ) 
		return -1;
	else
		return( MessagesReceived->Header.tag );
}
/*}}}*/

/*{{{  _probe*/
/*
 *------------------------------------------------------------------------
 *
 *      _probe :
 *		Checks if any message has arrived from a specified source
 *	Parameters :
 *		NodeId, TaskId: id of source, -1 indicates any source
 *
 *      Return value :
 *              1 if any messages have arrived, 0 otherwise
 *
 *------------------------------------------------------------------------
 */
int _probe( int NodeId, int TaskId, int tag )
{
	MESSAGE *tempMessage;
	PRIORITY_TYPE CurrentPri;

	CurrentPri = ProcSetPri( High );
	FORALL( MessagesReceived, tempMessage ){
		if ( MATCHES( NodeId, tempMessage->SourceNode ) &&
		     MATCHES( tag, tempMessage->Header.tag ) &&
		     MATCHES( TaskId, tempMessage->SourceTask )){
			ProcSetPri( CurrentPri );
			return(1);
		}
	}
	ProcSetPri( CurrentPri );
	return(0);
}
/*}}}*/

_kill_task( int NodeId, int TaskId )
{
	MESSAGE_HEADER header;
	Channel *VChan;
	int	index;
	PRIORITY_TYPE CurrentPri;

	CurrentPri = ProcSetPri( High );
	header.MessageLength = 0;
	header.tag = 0;
	OutChannel( VChan, NodeId, TaskId );
	header.flag = EXIT_MESSAGE;
	_SemWait( &SenderSemaphore[NodeId][TaskId] );
	_VChanOut( VChan, &header, FIXUP_LEN(sizeof(header)));
	_SemSignal( &SenderSemaphore[NodeId][TaskId] );
	ProcSetPri( CurrentPri );
	return(SUCCESS);
}


_kill_receiver( int NodeId, int TaskId )
{
	MESSAGE_HEADER header;
	Channel *VChan;
	int	index;
	PRIORITY_TYPE CurrentPri;

	CurrentPri = ProcSetPri( High );
	header.MessageLength = 0;
	header.tag = 0;
	OutChannel( VChan, NodeId, TaskId );
	header.flag = STOP_MESSAGE;
	_SemWait( &SenderSemaphore[NodeId][TaskId] );
	_VChanOut( VChan, &header, FIXUP_LEN(sizeof(header)));
	_SemSignal( &SenderSemaphore[NodeId][TaskId] );
	ProcSetPri( CurrentPri );
	return(SUCCESS);
}

ResetAllChannels()
{
	int i;
	i = 0;
	if ( VirtualChannelList == NULL )return;
	while ( VirtualChannelList[i] != NULL ){
		ResetChannel(VirtualChannelList[i] );
		i++;
	}
}


ThreadRunHigh( int *WorkSpace, void fn() )
{
	PRIORITY_TYPE CurrentPri;

	CurrentPri = ProcSetPri( High );
	ThreadRun( WorkSpace, fn );
	ProcSetPri( CurrentPri );
}

