#
# $Id: util.tcl,v 2.310 1994/07/18 18:18:43 kohl Exp $
#

#
# XPVM TCL Utility Procs
#

proc scrollInit \
	{ canvas sb canvas_size win_size vvar orient loc incr } \
{
	#debug_proc scrollInit entry

	global $vvar

	if { $loc == "low" } \
		{ set first 0 } \

	elseif { $loc == "center" } \
	{
		set first \
			[ expr ( ($canvas_size / 2) - ($win_size / 2) ) / $incr ]
	} \

	elseif { $loc == "high" } \
		{ set first [ expr ( $canvas_size - $win_size ) / $incr ] }

	scrollSet $sb \
		[ expr $canvas_size / $incr ] [ expr $win_size / $incr ] $first

	if { $orient == "HORIZ" } \
	{
		$canvas xview $first

		set $vvar $first
	} \

	else \
	{
		$canvas yview $first

		set $vvar $first
	}

	#debug_proc scrollInit exit
}

proc scrollSet { sb canvas_size win_size first } \
{
	#debug_proc scrollSet entry

	set last [ expr $first + ( $win_size - 1 ) ]

	$sb set $canvas_size $win_size $first $last

	#debug_proc scrollSet exit
}

proc scrollViews \
	{ canvas sb canvas_size win_size vvar orient view incr value } \
{
	#debug_proc scrollViews entry

	global st_active
	global ut_active
	global st_cwidth
	global ut_cwidth
	global timex

	scrollCanvas $canvas $sb $canvas_size $win_size $vvar $orient \
		$incr $value

	if { $view == "ST" && $ut_active == "TRUE" } \
	{
		set utxv [ expr $value - \
			( $value * ( $ut_cwidth - $st_cwidth) \
				/ ( $timex - $st_cwidth ) ) ]

		scrollCanvas \
			.utilization.frame.canvas .utilization.frame.horiz_sb \
			timex ut_cwidth ut_xview HORIZ $incr $utxv
	} \

	elseif { $view == "UT" && $st_active == "TRUE" } \
	{
		set stxv [ expr $value + \
			( $value * ( $ut_cwidth - $st_cwidth) \
				/ ( $timex - $ut_cwidth ) ) ]

		scrollCanvas \
			.xpvm.space_time.canvas .xpvm.space_time.horiz_sb \
			timex st_cwidth st_xview HORIZ $incr $stxv
	}

	#debug_proc scrollViews exit
}

proc scrollTwoCanvases \
	{ c1 c2 sb csize wsize vvar orient incr value } \
{
	#debug_proc scrollTwoCanvases entry

	scrollCanvas $c1 $sb $csize $wsize $vvar $orient $incr $value

	scrollCanvas $c2 $sb $csize $wsize $vvar $orient $incr $value

	#debug_proc scrollTwoCanvases exit
}

proc scrollCanvas \
	{ canvas sb canvas_size win_size vvar orient incr value } \
{
	#debug_proc scrollCanvas entry

	global $canvas_size
	global $win_size
	global $vvar

	global frame_border

	#
	# Limit Scroll Value
	#

	set csize [ expr $$canvas_size ]

	set wsize [ expr $$win_size ]

	if { $value < 0 } { set value 0 }

	set maxvalue [ expr $csize - $wsize + (2 * $frame_border) \
		+ ($incr - 1) ]

	if { $maxvalue > 0 } \
	{
		if { [ expr $value * $incr ] > $maxvalue } \
			{ set value [ expr $maxvalue / $incr ] }
	} \

	else \
		{ set value 0 }

	#
	# Set Scrollbar
	#

	scrollSet $sb [ expr $csize / $incr ] [ expr $wsize / $incr ] $value

	#
	# Scroll Canvas View
	#

	if { $orient == "HORIZ" } \
	{
		$canvas xview $value

		set $vvar $value
	} \

	else \
	{
		$canvas yview $value

		set $vvar $value
	}

	#debug_proc scrollCanvas exit
}

proc makeMenu { name cmd info_list close prefix } \
{
	#debug_proc makeMenu entry

	global active_fg_color
	global selector_color
	global frame_border
	global fg_color

	frame $name -bd $frame_border -relief raised

	set root [ rootName $name ]

	set mil "$root\_list"

	global $mil

	set LAST none

	set width 0

	foreach i $info_list \
	{
		set label	[ lindex $i 0 ]
		set type	[ lindex $i 1 ]
		set index	[ lindex $i 2 ]

		#
		# Create Button / Label
		#

		set fix_label [ strip_label $label ]

		if { $type == "host" || $type == "checkbutton" } \
		{
			#
			# Get State From Info List
			#

			set state	[ lindex $i 3 ]

			#
			# Create checkbutton state var
			#

			if { $prefix == "none" } \
				{ set prefix $root }

			if { $type == "host" } \
				{ set menuvar "$prefix\_state_$fix_label\_$index" } \

			else \
				{ set menuvar "$prefix\_state_$fix_label" }

			global $menuvar

			if { $state == "ON" } \
				{ set $menuvar ON } \

			else \
				{ set $menuvar OFF }

			#
			# Extract Host Stuff
			#

			if { $type == "host" } \
			{
				set icon	[ lindex $i 4 ]
				set iwt		[ lindex $i 5 ]
				set iht		[ lindex $i 6 ]

				addNetworkHost $label $icon $iwt $iht $state $menuvar
			}

			#
			# Set checkbutton command / create
			#

			set command "$cmd $index \"$label\" $menuvar FALSE"

			set NAME "$name.ckbutt_$fix_label"

			checkbutton $NAME -text $label -command $command \
				-bd 1 -relief raised -anchor nw \
				-onvalue ON -offvalue OFF -variable $menuvar \
				-selector $selector_color -foreground $fg_color \
				-activeforeground $active_fg_color
		} \

		elseif { $type == "command" } \
		{
			set command "$cmd $index \"$label\" none FALSE"

			set NAME "$name.butt_$fix_label"

			button $NAME -text $label -command $command \
				-bd 1 -relief raised \
				-foreground $fg_color -activeforeground $active_fg_color
		} \

		elseif { $type == "submenu" } \
		{
			set focus_entry [ lindex $i 3 ]
			set subs		[ lindex $i 4 ]

			set NAME "$name.butt_$fix_label"

			set command "raiseSubMenu $index $name $NAME \
				$focus_entry $subs"

			button $NAME -text $label -command $command \
				-bd 1 -relief raised -foreground $fg_color \
				-activeforeground $active_fg_color
		} \

		elseif { $type == "exchange" } \
		{
			set button	[ lindex $i 3 ]
			set parent	[ lindex $i 4 ]
			set subs	[ lindex $i 5 ]

			set command \
				"exchangeMenu $index $name $button $parent $subs"

			set NAME "$name.butt_$fix_label"

			button $NAME -text $label -command $command \
				-bd 1 -relief raised -foreground $fg_color \
				-activeforeground $active_fg_color
		} \

		else \
		{
			set NAME "$name.label_$fix_label"

			label $NAME -text $label -foreground $fg_color
		}

		#
		# Place Button / Label on Menu Frame
		#

		if { $LAST != "none" } \
			{ place $NAME -x $frame_border -y [ below $LAST ] } \

		else \
			{ place $NAME -x $frame_border -y $frame_border }

		update

		#
		# Process String Length & Button Width
		#

		set wt [ winfo width $NAME ]

		if { $wt > $width } \
			{ set width $wt }

		#
		# Save Button Name
		#

		set LAST $NAME

		lappend $mil $NAME
	}

	#
	# Create "Done" Button
	#

	set NAME $name.butt_Done

	if { "$close" == "lower" } \
		{ set cmd "lowerMenu $name" } \

	else \
		{ set cmd "$close" }

	button $NAME -text "Done" \
		-bd 1 -relief raised -command $cmd \
		-foreground $fg_color -activeforeground $active_fg_color

	if { $LAST != "none" } \
		{ place $NAME -x $frame_border -y [ below $LAST ] } \

	else \
		{ place $NAME -x $frame_border -y $frame_border }

	lappend $mil $NAME

	update

	#
	# Set Menu Frame Size
	#

	set menu_height [ expr [ below $NAME ] + 4 ]

	set menu_width [ expr $width + 8 ]

	$name configure -width $menu_width -height $menu_height

	place forget $name

	#
	# Adjust Menu Button Widths
	#

	foreach i [ expr $$mil ] \
		{ place $i -width $width }

	#debug_proc makeMenu exit
}

proc raiseMenu { name button parent subs } \
{
	#debug_proc raiseMenu entry

	foreach i $subs \
		{ place forget $i }

	set ckm [ winfo ismapped $name ]

	if { $ckm == 1 } \
		{ place forget $name } \

	else \
		{ placeMenu $name $button $parent }

	#debug_proc raiseMenu exit
}

proc placeMenu { name button parent } \
{
	#debug_proc placeMenu entry

	set px [ winfo rootx $parent ]
	set py [ winfo rooty $parent ]

	set x [ expr [ winfo rootx $button ] - $px ]

	set y [ expr [ winfo rooty $button ] \
		+ [ winfo height $button ] - $py ]

	place $name -x $x -y $y

	update

	check_in_main $name $x $y

	#debug_proc placeMenu exit
}

proc replaceMenu { name button parent } \
{
	#debug_proc replaceMenu entry

	set ckup [ winfo ismapped $name ]

	if { $ckup == 1 } \
		{ placeMenu $name $button $parent }

	#debug_proc replaceMenu exit
}

proc raiseSubMenu { submenu menu button focus_entry subs } \
{
	#debug_proc raiseSubMenu entry

	foreach i $subs \
		{ place forget $i }

	set ckup [ winfo ismapped $submenu ]

	if { $ckup == 1 } \
		{ place forget $submenu } \

	else \
		{ placeSubMenu $submenu $menu $button $focus_entry }

	#debug_proc raiseSubMenu exit
}

proc placeSubMenu { submenu menu button focus_entry } \
{
	#debug_proc placeSubMenu entry

	set x [ winfo x $menu ]
	set y [ winfo y $menu ]

	set x [ expr $x + [ left $button ] ]
	set y [ expr $y + [ below $button ] ]

	place $submenu -x $x -y $y

	update

	check_in_main $submenu $x $y

	if { $focus_entry != "none" } \
		{ focus $focus_entry }

	#debug_proc placeSubMenu exit
}

proc replaceSubMenu { submenu menu button focus_entry } \
{
	#debug_proc replaceSubMenu entry

	set ckup [ winfo ismapped $submenu ]

	if { $ckup == 1 } \
		{ placeSubMenu $submenu $menu $button $focus_entry }

	#debug_proc replaceSubMenu exit
}

proc exchangeMenu { submenu menu button parent subs } \
{
	#debug_proc exchangeMenu entry

	set cksm [ winfo ismapped $submenu ]

	if { $cksm == 1 } \
	{
		place forget $submenu

		raiseMenu $menu $button $parent $subs
	} \

	else \
	{
		place forget $menu

		raiseMenu $submenu $button $parent $subs
	}

	#debug_proc exchangeMenu exit
}

proc check_in_main { name x y } \
{
	#debug_proc check_in_main entry

	global main_height
	global main_width

	set wt [ winfo width $name ]

	set ckx [ expr $x + $wt ]

	if { $ckx > $main_width } \
	{
		set x [ expr $main_width - $wt ]

		if { $x < 0 } { set x 0 }

		place $name -x $x
	}

	set ht [ winfo height $name ]

	set cky [ expr $y + $ht ]

	if { $cky > $main_height } \
	{
		set y [ expr $main_height - $ht ]

		if { $y < 0 } { set y 0 }

		place $name -y $y
	}

	#debug_proc check_in_main exit
}

proc lowerMenu { name } \
{
	#debug_proc lowerMenu entry

	place forget $name

	#debug_proc lowerMenu exit
}

proc refresh_tasks { type } \
{
	#debug_proc refresh_tasks entry

	global frame_border

	global $type\_tasklist
	global $type\_lasty

	#
	# Delete Any Old Task Buttons
	#

	foreach tb [ expr $$type\_tasklist ] \
	{
		if { "$tb" != "none" } \
		{
			set bid [ lindex $tb 3 ]

			.xpvm.$type.canvas delete $bid
		}
	}

	set $type\_tasklist none

	set $type\_lasty $frame_border

	#
	# Get New Task List & Create Buttons
	#

	pvm_get_tasks $type

	#debug_proc refresh_tasks exit
}

proc add_list_task { type host name tid } \
{
	#debug_proc add_list_task entry

	global $type\_tasklist
	global $type\_cheight
	global $type\_lasty
	global frame_border
	global row_height
	global fg_color

	set str "$host:$name, TID=$tid"

	set id [ .xpvm.$type.canvas create text \
		$frame_border [ expr $$type\_lasty ] \
		-text $str -anchor nw -fill $fg_color ]

	set $type\_lasty [ expr [ expr $$type\_lasty ] + $row_height ]

	scrollSet .xpvm.$type.vert_sb \
		[ expr $$type\_lasty ] [ expr $$type\_cheight ] 0

	set t [ list $host $name $tid $id ]

	set $type\_tasklist [ linsert [ expr $$type\_tasklist ] 0 $t ] 

	#debug_proc add_list_task exit
}

proc trace_controls { status } \
{
	#debug_proc trace_controls entry

	global trace_control_status

	set CTRL .xpvm.controls

	# Un-Highlight Previous Button

	if { $trace_control_status != "none" } \
		{ $CTRL.$trace_control_status configure -relief raised }

	# Highlight New Button

	$CTRL.$status configure -relief sunken

	set trace_control_status $status

	update

	#debug_proc trace_controls exit
}

proc setMsg { text } \
{
	#debug_proc setMsg entry

	.xpvm.message configure -text "Status:   $text"

	update

	#debug_proc setMsg exit
}

proc rootName { path } \
{
	#debug_proc rootName entry

	set tmp [ split $path . ]

	set num [ llength $tmp ]

	set root [ lindex $tmp [ expr $num - 1 ] ]

	return $root

	#debug_proc rootName exit
}

proc x_of_time { sec usec scale } \
{
	#debug_proc x_of_time entry

	global frame_border

	return [ expr ( ( ( $sec * 1000000 ) + $usec ) / $scale ) \
		+ $frame_border ]

	#debug_proc x_of_time exit
}

proc above { win } \
{
	#debug_proc above entry

	set y [ winfo y $win ]

	return $y

	#debug_proc above exit
}

proc below { win } \
{
	#debug_proc below entry

	set y [ expr [ winfo y $win ] + [ winfo height $win ] ]

	return $y

	#debug_proc below exit
}

proc left { win } \
{
	#debug_proc left entry

	set x [ winfo x $win ]

	return $x

	#debug_proc left exit
}

proc right { win } \
{
	#debug_proc right entry

	set x [ expr [ winfo x $win ] + [ winfo width $win ] ]

	return $x

	#debug_proc right exit
}

proc set_x_coords { C id x1 x2 } \
{
	#debug_proc set_x_coords entry

	set coords [ $C coords $id ]

	if { $x1 == -1 } { set x1 [ lindex $coords 0 ] }

	if { $x2 == -1 } { set x2 [ lindex $coords 2 ] }

	set y1 [ lindex $coords 1 ]
	set y2 [ lindex $coords 3 ]

	$C coords $id $x1 $y1 $x2 $y2

	#debug_proc set_x_coords exit
}

proc set_y_coords { C id y1 y2 } \
{
	#debug_proc set_y_coords entry

	set coords [ $C coords $id ]

	if { $y1 == -1 } { set y1 [ lindex $coords 1 ] }

	if { $y2 == -1 } { set y2 [ lindex $coords 3 ] }

	set x1 [ lindex $coords 0 ]
	set x2 [ lindex $coords 2 ]

	$C coords $id $x1 $y1 $x2 $y2

	#debug_proc set_y_coords exit
}

proc get_user_name { } \
{
	#debug_proc get_user_name entry

	set home [ glob ~ ]

	set len [ string length $home ]

	set slash_index [ string last / $home ]

	set user_index [ expr $slash_index + 1 ]

	set user [ string range $home $user_index $len ]

	return $user

	#debug_proc get_user_name exit
}

proc helpSelect { index label menuvar setflag } \
{
	#debug_proc helpSelect entry

	global active_fg_color
	global min_help_height
	global min_help_width
	global scroll_width
	global border_space
	global frame_border
	global row_height
	global col_width
	global fg_color
	global help_dir

	#
	# Check for Existing Help Window
	#
	
	set WIN .$index\_help_win

	set hwinck [ winfo exists $WIN ]

	if { $hwinck == 1 } \
	{
		set hwstate [ wm state $WIN ]

		if { $hwstate == "normal" } \
		{
			global $index\_visibility

			set vis [ expr $$index\_visibility ]

			if { $vis == "VisibilityUnobscured" } \
				{ wm withdraw $WIN } \
			
			else \
			{
				set $index\_visibility "VisibilityUnobscured"

				raise $WIN
			}
		} \
		
		else \
			{ wm deiconify $WIN }

		return
	}

	#
	# Get Help File
	#

	set fname $help_dir/$index.help

	set fileck [ file exists $fname ]

	if { $fileck != 1 } \
	{
		setMsg "Help File $fname Not Found"

		return
	}

	set fp [ open $fname ]

	set whole [ read $fp ]

	set title [ lindex $whole 0 ]

	set text [ lreplace $whole 0 0 ]

	close $fp

	#
	# Create Help Window
	#

	global help_stagger_index
	global $index\_canvas_height
	global $index\_canvas_width
	global $index\_visibility
	global $index\_cheight
	global $index\_cwidth
	global $index\_height
	global $index\_width

	set $index\_height [ expr $row_height * 30 ]
	set $index\_width [ expr $col_width * 40 ]

	set ht [ expr $$index\_height ]
	set wt [ expr $$index\_width ]

	set stagx [ expr 50 + ( $help_stagger_index * $frame_border * 2 ) ]
	set stagy [ expr 125 + ( $help_stagger_index * $border_space ) ]

	set help_stagger_index [ expr $help_stagger_index + 1 ]

	set hgeom "[expr $wt]x[expr $ht]+[expr $stagx]+[expr $stagy]"

	toplevel $WIN

	wm geometry $WIN $hgeom

	wm minsize $WIN $min_help_width $min_help_height

	set $index\_visibility "VisibilityUnobscured"

	set WIN_C $WIN.canvas

	set WIN_SBH $WIN.horiz_sb
	set WIN_SBV $WIN.vert_sb

	set $index\_cheight [ expr $$index\_height - $scroll_width \
		- (2 * $row_height) - (2 * $border_space) ]

	set $index\_cwidth [ expr $$index\_width - $scroll_width \
		- (2 * $border_space) ]

	label $WIN.title -text $title -foreground $fg_color

	set cht [ expr $row_height * [ llength $text ] ]

	set cwt [ expr $col_width * 80 ]

	canvas $WIN_C -bd $frame_border -relief sunken \
		-scrollincrement $row_height -confine 0 \
		-scrollregion "0 0 $cwt $cht"

	set $index\_canvas_height $cht

	set $index\_canvas_width $cwt

	set cmd "scrollCanvas $WIN_C $WIN_SBH \
		$index\_canvas_width $index\_cwidth $index\_xview \
		HORIZ $row_height"

	scrollbar $WIN_SBH -orient horiz -bd $frame_border -relief sunken \
		-width $scroll_width -command $cmd

	scrollInit $WIN_C $WIN_SBH $cwt [ expr $$index\_cwidth ] \
		$index\_xview HORIZ low $row_height

	set cmd "scrollCanvas $WIN_C $WIN_SBV \
		$index\_canvas_height $index\_cheight $index\_yview \
		VERT $row_height"

	scrollbar $WIN_SBV -orient vert -bd $frame_border -relief sunken \
		-width $scroll_width -command $cmd

	scrollInit $WIN_C $WIN_SBV $cht [ expr $$index\_cheight ] \
		$index\_yview VERT low $row_height

	button $WIN.close -text "Close" -command "wm withdraw $WIN" \
		-foreground $fg_color -activeforeground $active_fg_color

	layout_help_win $index

	bind $WIN <Configure> "resize_help_win $index"

	bind $WIN <Visibility> "help_visibility $index %s"

	#
	# Fill Canvas With Text
	#

	set ht $border_space
	set wt $border_space

	set cnt 0

	foreach line $text \
	{
		set lid "$WIN.line$cnt"

		set fix_line [ fix_help_line $line ]

		set wid [ $WIN_C create text $wt $ht -text $fix_line \
			-anchor nw -fill $fg_color ]

		set ht [ expr $ht + $row_height ]

		set cnt [ expr $cnt + 1 ]
	}

	#debug_proc helpSelect exit
}

proc help_visibility { index detail } \
{
	#debug_proc help_visibility entry

	global $index\_visibility

	set $index\_visibility $detail

	#debug_proc help_visibility exit
}

proc unmap_all { } \
{
	#debug_proc unmap_all entry

	global win_map_list

	set win_map_list ""

	set tmp_list ".xpvm"

	foreach w { .utilization .call_trace .task_output } \
	{
		set ckup [ winfo ismapped $w ]

		if { $ckup == 1 } \
			{ lappend tmp_list $w }
	}

	set tmp_list [ sort_map_list $tmp_list ]

	foreach w $tmp_list \
		{ wm iconify $w }

	set win_map_list $tmp_list

	#debug_proc unmap_all exit
}

proc remap_all { } \
{
	#debug_proc remap_all entry

	global win_map_list

	set save_list $win_map_list

	set win_map_list ""

	foreach w $save_list \
		{ wm deiconify $w }

	foreach w $save_list \
		{ raise $w }

	#debug_proc remap_all exit
}

proc sort_map_list { wlist } \
{
	#debug_proc sort_map_list entry

	set num [ llength $wlist ]

	set done 0

	while { $done == 0 } \
	{
		set done 1

		for { set i 0 } { $i < $num - 1 } { set i [ expr $i + 1 ] } \
		{
			set ip1 [ expr $i + 1 ]

			set w [ lindex $wlist $i ]
			set t [ lindex $wlist $ip1 ]

			set tw [ top_window $w $t ]

			if { $tw == $w } \
			{
				set wlist [ lreplace $wlist $i $ip1 $t $w ]

				set done 0
			}
		}
	}

	#debug_proc sort_map_list exit

	return $wlist
}

proc top_window { w1 w2 } \
{
	#debug_proc top_window entry

	set tw [ do_top_window $w1 $w2 ]

	if { $tw == "" } \
		{ set tw [ do_top_window $w2 $w1 ] }

	#debug_proc top_window exit

	return $tw
}

proc do_top_window { w1 w2 } \
{
	#debug_proc do_top_window entry

	set x1 [ winfo x $w1 ]
	set y1 [ winfo y $w1 ]

	set x2 [ winfo x $w2 ]
	set y2 [ winfo y $w2 ]

	set ht1 [ winfo height $w1 ]
	set wt1 [ winfo width $w1 ]

	set ht2 [ winfo height $w2 ]
	set wt2 [ winfo width $w2 ]

	if { $x1 <= $x2 && $x2 <= $x1 + $wt1 \
		&& $y1 <= $y2 && $y2 <= $y1 + $ht1 } \
	{
		#
		# ( x2, y2 ) inside w1
		#

		set tw [ winfo containing $x2 $y2 ]
	} \

	elseif { $x1 <= $x2 + $wt2 && $x2 + $wt2 <= $x1 + $wt1 \
		&& $y1 <= $y2 && $y2 <= $y1 + $ht1 } \
	{
		#
		# ( x2 + wt2, y2 ) inside w1
		#

		set tw [ winfo containing [ expr $x2 + $wt2 ] $y2 ]
	} \

	elseif { $x1 <= $x2 && $x2 <= $x1 + $wt1 \
		&& $y1 <= $y2 + $ht2 && $y2 + $ht2 <= $y1 + $ht1 } \
	{
		#
		# ( x2, y2 + ht2 ) inside w1
		#

		set tw [ winfo containing $x2 [ expr $y2 + $ht2 ] ]
	} \

	elseif { $x1 <= $x2 + $wt2 && $x2 + $wt2 <= $x1 + $wt1 \
		&& $y1 <= $y2 + $ht2 && $y2 + $ht2 <= $y1 + $ht1 } \
	{
		#
		# ( x2 + wt2, y2 + ht2 ) inside w1
		#

		set tw [ winfo containing \
			[ expr $x2 + $wt2 ] [ expr $y2 + $ht2 ] ]
	} \

	else \
		{ set tw "" }

	if { $tw != "" } \
		{ set tw [ winfo toplevel $tw ] }

	#debug_proc do_top_window exit

	return $tw
}

proc exit { } \
{
	#debug_proc exit entry

	pvm_quit

	#debug_proc exit exit
}

proc quit { } \
{
	#debug_proc quit entry

	pvm_quit

	#debug_proc quit exit
}

proc halt { } \
{
	#debug_proc halt entry

	pvm_halt

	#debug_proc halt exit
}

proc debug_proc { routine inout } \
{
	global proc_debug

	if { $proc_debug == "TRUE" } \
		{ puts "(proc debug: $routine{} $inout)" }
}

